/**
 * @file get_rtas_event.c
 * @brief librtasevent routines to parse RTAS events
 *
 * Copyright (C) 2005 IBM Corporation
 * Common Public License Version 1.0 (see COPYRIGHT)
 *
 * @author Nathan Fontenot <nfont@austin.ibm.com>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "librtasevent.h"
#include "rtas_event.h"

/**
 * cleanup_rtas_event
 * @brief free the structures related to a parsed rtas event
 *
 * @param re rtas_event pointer
 * @return 0 on success, !0 on failure
 */
int
cleanup_rtas_event(struct rtas_event *re)
{
    struct rtas_event_scn *res;

    if (re == NULL)
        return 0;

    res = re->event_scns;

    while (res != NULL) {
        struct rtas_event_scn *tmp = res;
        res = res->next;

        if (tmp->subscn != NULL)
            free(tmp->subscn);

        if (tmp->other_scns != NULL) {
            struct rtas_event_scn *tmp2, *tmp3;
            tmp2 = tmp->other_scns;
            while (tmp2 != NULL) {
		tmp3 = tmp2;
                tmp2 = tmp2->next;
                free(tmp3);
            }
        }

        free(tmp);
    }

    free(re);

    return 0;
}

/**
 * add_re_subscn
 * @brief Add a rtas_event_scn for a subsection of an existing section
 *
 * @param res rtas_event_scn pointer to section this is to be subsection of
 * @param scn start of the rtas event section to add
 * @param scn_id section id of section to be added
 * @return pointer to newly added rtas_event_scn
 */
static struct rtas_event_scn *
add_re_subscn(struct rtas_event_scn *res, char *scn, int scn_id)
{
    struct rtas_event_scn *new_res;

    new_res = (struct rtas_event_scn *)malloc(sizeof(*new_res));
    if (new_res == NULL) {
        errno = ENOMEM;
        return NULL;
    }

    new_res->next = NULL;
    new_res->data = scn;
    new_res->scn_id = scn_id;
    new_res->re = res->re;
    new_res->subscn = NULL;
    new_res->other_scns = NULL;

    res->subscn = new_res;

    return new_res;
}

/**
 * add_re_other_scn
 * @brief add a rtas_event_scn to the list of other sections
 *
 * @param res rtas_event_scn to add the new rtas section to
 * @param scn pointer to the beginning of the section to be added
 * @param scn_id id of the section to be added
 * @return pointer to the newly created rtas_event_scn
 */
static struct rtas_event_scn *
add_re_other_scn(struct rtas_event_scn *res, char *scn, int scn_id)
{
    struct rtas_event_scn *new_res, *tmp_res;

    new_res = (struct rtas_event_scn *)malloc(sizeof(*new_res));
    if (new_res == NULL) {
        errno = ENOMEM;
        return NULL;
    }

    new_res->next = NULL;
    new_res->data = scn;
    new_res->scn_id = scn_id;
    new_res->re = res->re;
    new_res->subscn = NULL;
    new_res->other_scns = NULL;

    if (res->other_scns == NULL) {
        res->other_scns = new_res;
    } 
    else {
        tmp_res = res->other_scns;
        while (tmp_res->next != NULL)
            tmp_res = tmp_res->next;

        tmp_res->next = new_res;
    }

    return new_res;
}

/**
 * add_re_scn
 * @brief Add a rtas event section to the section list
 *
 * @param re rtas_event pointer to add this section to
 * @param scn pointer to start of rtas event section
 * @param scn_id id of the section to be added
 * return pointer to newly created rtas_event_scn
 */
static struct rtas_event_scn *
add_re_scn(struct rtas_event *re, char *scn, int scn_id)
{
    struct rtas_event_scn *new_res, *tmp_res;

    new_res = (struct rtas_event_scn *)malloc(sizeof(*new_res));
    if (new_res == NULL) {
        errno = ENOMEM;
        return NULL;
    }

    new_res->next = NULL;
    new_res->data = scn;
    new_res->scn_id = scn_id;
    new_res->re = re;
    new_res->subscn = NULL;
    new_res->other_scns = NULL;

    if (re->event_scns == NULL) {
        re->event_scns = new_res;
    } 
    else {
        tmp_res = re->event_scns;
        while (tmp_res->next != NULL)
            tmp_res = tmp_res->next;

        tmp_res->next = new_res;
    }

    return new_res;
}

/**
 * re_scn_id
 * @brief Convert the two character section id into an internal identifier
 *
 * @param shdr rtas_v6_scn_hdr pointer
 * @return section identifier on success, -1 on failure
 */
static int
re_scn_id(struct rtas_v6_scn_hdr *shdr) 
{
    if (strncmp(shdr->id, RTAS_DUMP_SCN_ID, 2) == 0)
        return RTAS_DUMP_SCN;
    if (strncmp(shdr->id, RTAS_EPOW_SCN_ID, 2) == 0)
        return RTAS_EPOW_SCN;
    if (strncmp(shdr->id, RTAS_IO_SCN_ID, 2) == 0)
        return RTAS_IO_SCN;
    if (strncmp(shdr->id, RTAS_LRI_SCN_ID, 2) == 0)
        return RTAS_LRI_SCN;
    if (strncmp(shdr->id, RTAS_MTMS_SCN_ID, 2) == 0)
        return RTAS_MTMS_SCN;
    if (strncmp(shdr->id, RTAS_PSRC_SCN_ID, 2) == 0)
        return RTAS_PSRC_SCN;
    if (strncmp(shdr->id, RTAS_SSRC_SCN_ID, 2) == 0)
        return RTAS_SSRC_SCN;

    return -1;
}

/**
 * _get_re_scn
 * @brief the real work for retreiving rtas event sections
 *
 * @param res rtas_event_scn pointer at which to start search
 * @param scn_id section id to search for
 * @return rtas_event_scn on success, NULL on failure
 */
static struct rtas_event_scn *
_get_re_scn(struct rtas_event_scn *res, int scn_id)
{
    for ( ; res != NULL; res = res->next) {
        if (res->scn_id == scn_id)
            return res;

	if (res->subscn != NULL) {
            if (res->subscn->scn_id == scn_id)
                return res->subscn;
        }

        if (res->other_scns != NULL) {
            struct rtas_event_scn *tmp;

            tmp = _get_re_scn(res->other_scns, scn_id);
            if (tmp != NULL)
                return tmp;
        }
    }

    return NULL;
}

/**
 * get_re_scn
 * @brief find the specified section on the list of sections
 *
 * NOTE: this function has been split just so we can have common place
 * to check for NULL 're' pointers (no sense checking everywhere we
 * call this). 
 *
 * @param re rtas_event pointer
 * @param scn_id id of the section to find
 * @return rtas_event_scn on success, NULL on failure
 */
struct rtas_event_scn *
get_re_scn(struct rtas_event *re, int scn_id)
{
    if (re == NULL) {
        errno = EFAULT;
        return NULL;
    }

    return _get_re_scn(re->event_scns, scn_id);
}

/**
 * parse_v6_src_scn
 * @brief parse a version 6 rtas SRC section
 *
 * @param re rtas_event pointer
 * @param src_start pointer to beginning of SRC section
 * @return 0 on success, !0 on failure
 */
static int
parse_v6_src_scn(struct rtas_event *re, char *src_start)
{
    struct rtas_event_scn *srcres, *frures;
    struct rtas_v6_src_scn *src;
    struct rtas_v6_src_subscn *srcsub;
    struct rtas_v6_fru_scn *fru;
    struct rtas_v6_fru_subscn *frusub;
    int total_len, srcsub_len;

    src = (struct rtas_v6_src_scn *)src_start;

    srcres = add_re_scn(re, (char *)src, re_scn_id(&src->shdr));
    if (srcres == NULL) {
        cleanup_rtas_event(re);
        return -1;
    }

    if (src->subscns == 0)
        return 0;

    srcsub = (struct rtas_v6_src_subscn *)((char *)src + sizeof(*src));
    if (add_re_subscn(srcres, (char *)srcsub, RTAS_SRC_SUBSCN) == NULL) {
        cleanup_rtas_event(re);
        return -1;
    }

    srcsub_len = srcsub->subscn_length * 4; /*get number of bytes */
    total_len = sizeof(*srcsub);

    do {
        fru = (struct rtas_v6_fru_scn *)((char *)srcsub + total_len);
        frures = add_re_other_scn(srcres, (char *)fru, RTAS_FRU_SCN);
        if (frures == NULL) {
            cleanup_rtas_event(re);
            return -1;
        }

        if (fru->subscn_included) {
            frusub = (struct rtas_v6_fru_subscn *)
                     ((char *)fru + sizeof(*fru) + fru->loc_code_length);
            if (add_re_subscn(frures, (char *)frusub, 
			      RTAS_FRU_SUBSCN) == NULL) {
                cleanup_rtas_event(re);
                return -1;
            }
        }

        total_len += fru->length;
    } while (total_len < srcsub_len);

    return 0;
}

/**
 * parse_v6_rtas_event
 * @brief parse a version 6 RTAS event
 *
 * @param re rtas_event pointer
 * @return rtas_event pointer on success, NULL on failure
 */
static struct rtas_event *
parse_v6_rtas_event(struct rtas_event *re)
{
    struct rtas_event_hdr *re_hdr;
    struct rtas_event_exthdr *rex_hdr;
    struct rtas_v6_ibm_platform_event *pe_hdr;
    struct rtas_v6_scn_hdr *shdr;
    int total_len;
    
    re_hdr = (struct rtas_event_hdr *)re->buffer;
    total_len = sizeof(*re_hdr);

    rex_hdr = (struct rtas_event_exthdr *)(re->buffer + total_len);
    total_len += sizeof(*rex_hdr);

    pe_hdr = (struct rtas_v6_ibm_platform_event *)(re->buffer + total_len);
    total_len += sizeof(*pe_hdr);

    if (re->length < total_len) {
        cleanup_rtas_event(re);
        errno = EFAULT;
	return NULL;
    }

    /* Sanity Check for "IBM" string */
    if (strncmp(pe_hdr->pe_ibm, "IBM", 3) != 0) {
        cleanup_rtas_event(re);
        errno = EFAULT;
        return NULL;
    }

    if (add_re_scn(re, (char *)&pe_hdr->pe_main_a, RTAS_MAIN_A_SCN) == NULL) {
        cleanup_rtas_event(re);
        return NULL;
    }

    if (add_re_scn(re, (char *)&pe_hdr->pe_main_b, RTAS_MAIN_B_SCN) == NULL) {
        cleanup_rtas_event(re);
        return NULL;
    }

    while (total_len < re->length) {
        int scn_id;
        shdr = (struct rtas_v6_scn_hdr *)(re->buffer + total_len);

        scn_id = re_scn_id(shdr);

        if ((scn_id == RTAS_PSRC_SCN) || (scn_id == RTAS_SSRC_SCN)) {
            parse_v6_src_scn(re, (char *)shdr);
        } else {
            if (add_re_scn(re, (char *)shdr, re_scn_id(shdr)) == NULL) {
                cleanup_rtas_event(re);
                return NULL;
            }
        }

        total_len += shdr->length;
    }

    return re;
}

/**
 * parse_rtas_event
 * @brief parse an rtas event creating a populated rtas_event structure
 *
 * @param buf buffer containing the binary RTAS event
 * @param buflen length of the buffer 'buf'
 * @return pointer to rtas_event
 */
struct rtas_event *
parse_rtas_event(char *buf, int buflen)
{
    struct rtas_event *re;
    struct rtas_event_hdr *re_hdr;
    struct rtas_event_exthdr *rex_hdr;
    int parsed_len;
    int scn_len;
    int event_len;
    int scn_type;

    re = malloc(sizeof(*re));
    if (re == NULL) {
        errno = ENOMEM;
        return NULL;
    }

    re->buffer = buf;
    re->event_scns = NULL;
    re->event_no = -1;

    re_hdr = (struct rtas_event_hdr *)buf;

    parsed_len = sizeof(*re_hdr);
    if (buflen < parsed_len) {
        cleanup_rtas_event(re);
        errno = EFAULT;
	return NULL;
    }

    re->length = re_hdr->ext_log_length + sizeof(*re_hdr);
    if (re->length > buflen) {
        cleanup_rtas_event(re);
        return NULL;
    }

    if (add_re_scn(re, (char *)re_hdr, RTAS_EVENT_HDR) == NULL) {
        cleanup_rtas_event(re);
        return NULL;
    }

    re->version = re_hdr->version;

    if (re_hdr->extended == 0)
        return re;
    
    rex_hdr = (struct rtas_event_exthdr *)(buf + parsed_len);
    parsed_len += sizeof(*rex_hdr);
    if (buflen < parsed_len) {
        cleanup_rtas_event(re);
        errno = EFAULT;
	return NULL;
    }

    if (add_re_scn(re, (char *)rex_hdr, RTAS_EVENT_EXT_HDR) == NULL) {
        cleanup_rtas_event(re);
        return NULL;
    }

    if (re_hdr->version == 6)
	return parse_v6_rtas_event(re);

    event_len = re_hdr->ext_log_length + sizeof(*re_hdr);

    switch (rex_hdr->format_type) {
        case RTAS_EXTHDR_FMT_CPU:
            scn_type = RTAS_CPU_SCN;
            scn_len = sizeof(struct rtas_cpu_scn);
	    break;

        case RTAS_EXTHDR_FMT_EPOW:
            scn_type = RTAS_EPOW_SCN;
            scn_len = sizeof(struct rtas_epow_scn);
            break;

        case RTAS_EXTHDR_FMT_IBM_DIAG:
            scn_type= RTAS_IBM_DIAG_SCN;
            scn_len = sizeof(struct rtas_ibm_diag_scn); 
            break;

        case RTAS_EXTHDR_FMT_IO:
            scn_type = RTAS_IO_SCN;
            scn_len = sizeof(struct rtas_io_scn);
            break;

        case RTAS_EXTHDR_FMT_MEMORY:
            scn_type = RTAS_MEM_SCN;
            scn_len = sizeof(struct rtas_mem_scn);
            break;
	
        case RTAS_EXTHDR_FMT_POST:
            scn_type = RTAS_POST_SCN;
            scn_len = sizeof(struct rtas_post_scn);
            break;

        case RTAS_EXTHDR_FMT_IBM_SP:
            scn_type = RTAS_IBM_SP_SCN;
            scn_len = sizeof(struct rtas_ibmsp_scn);
	    break;

        case RTAS_EXTHDR_FMT_VEND_SPECIFIC_1:
        case RTAS_EXTHDR_FMT_VEND_SPECIFIC_2:
            scn_type = RTAS_VEND_SPECIFIC_SCN;
            /* Make sure we don't try to parse any further */
            scn_len = event_len;
            break;
            
        default:
            cleanup_rtas_event(re);
            errno = EFAULT;
            return NULL;
            break;
    }

    if (add_re_scn(re, (char *)(buf + parsed_len), scn_type) == NULL) {
        cleanup_rtas_event(re);
        return NULL;
    }

    parsed_len += scn_len;

    /* Look for any additional Vendor Error Logs */
    if (parsed_len < event_len) {
        if (add_re_scn(re, (char *)(buf + parsed_len), 
                       RTAS_VEND_SPECIFIC_SCN) == NULL) {
            cleanup_rtas_event(re);
            return NULL;
        }
    }

    return re;
}

/**
 * rtas_get_vend_specific
 * @brief retrive a vendor specific section of the RTAS event
 *
 * @param re parsed rtas event
 * @return reference to a rtas_event_scn on success, NULL on failure
 */
struct rtas_event_scn *
rtas_get_vend_specific_scn(struct rtas_event *re)
{
    return get_re_scn(re, RTAS_VEND_SPECIFIC_SCN);
}
