/**
 * @file rtas_lri.c
 * @brief RTAS Logical Resource Identification (LRI) Data routines
 *
 * Copyright (C) 2005 IBM Corporation
 * Common Public License Version 1.0 (see COPYRIGHT)
 *
 * @author Nathan Fontenot <nfont@austin.ibm.com>
 */

#include <stdio.h>
#include <errno.h>

#include "librtasevent.h"
#include "rtas_event.h"

/**
 * rtas_get_lri_scn
 * @brief Retrieve the Logical Resource ID (LRI) section of the RTAS Event
 *
 * @param re rtas_event pointer
 * @return rtas_event_scn pointer for lri section
 */
struct rtas_event_scn *
rtas_get_lri_scn(struct rtas_event *re)
{
    return get_re_scn(re, RTAS_LRI_SCN);
}

/**
 * print_re_lriscn
 * @brief print the contents of a LRI section
 *
 * @param res rtas_event_scn pointer for lri section
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_re_lri_scn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_lri_scn *lri;
    int len = 0;

    if ((res->scn_id != RTAS_LRI_SCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    lri = (struct rtas_v6_lri_scn *)res->data;

    len += print_v6_scn_hdr("Logical Resource Identification", 
                            &lri->shdr, verbosity);
    
    len += rtas_print(PRNT_FMT" ", "Resource Type:", lri->resource);
    switch (lri->resource) {
        case 0x10:
	    len += rtas_print("(Processor)\n" PRNT_FMT_R, "CPU ID:", 
                              lri->lri_cpu_id);
	    break;
            
        case 0x11:
	    len += rtas_print("(Shared Processor)\n" PRNT_FMT_R, 
                              "Entitled Capacity:", lri->capacity);
	    break;

        case 0x40: 
            len += rtas_print("(Memory Page)\n" PRNT_FMT_ADDR, 
                              "Logical Address:", lri->lri_mem_addr_hi,
                              lri->lri_mem_addr_lo);
	    break;

        case 0x41:
	    len += rtas_print("(Memory LMB)\n" PRNT_FMT_R, 
                              "DRC Index:", lri->lri_drc_index);
	    break;

	default:
	    len += rtas_print("(Unknown Resource)\n");
	    break;
    }
    
    len += rtas_print("\n");
    return len;
}
