/**
 * @file rtas_srcfru.c
 * @brief RTAS SRC/FRU sections routines
 *
 * Copyright (C) 2005 IBM Corporation
 * Common Public License Version 1.0 (see COPYRIGHT)
 *
 * @author Nathan Fontenot <nfont@austin.ibm.com>
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "librtasevent.h"
#include "rtas_event.h"

/* rtas_src_codes contains data used by these routines */ 
#include "rtas_src_codes.c"

/**
 * rtas_get_src_scn
 * @brief retrieve the RTAS src section for a RTAS event
 *
 * @param re rtas_event pointer
 * @return rtas_event_scn pointer for a SRC section
 */ 
struct rtas_event_scn *
rtas_get_src_scn(struct rtas_event *re)
{
    return get_re_scn(re, RTAS_PSRC_SCN);
}

/**
 * print_re_fru_subscn
 * @brief print the contents of a FRU subsection
 *
 * @param res rtas_event_scn pointer for the FRU subsection
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_re_fru_subscn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_fru_subscn *fru;
    int i, len = 0;

    if ((res->scn_id != RTAS_FRU_SUBSCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    fru = (struct rtas_v6_fru_subscn *)res->data;

    if (verbosity >= 2) {
        len += print_scn_title("FRU Sub-Section");
        len += rtas_print("%-20s      %c%c    "PRNT_FMT_R, 
                       "Section ID:", fru->id[0], fru->id[1], 
                       "Section Length:", fru->length);
    }

    if (fru->component) {
        len += rtas_print(PRNT_FMT" ", "Failing Component:", fru->component);
        switch(fru->component) {
            case 0x1:
                len += rtas_print("(\"normal\" hardware FRU)\n");
                break;
                
            case 0x2:
                len += rtas_print("(Code FRU)\n");
                break;
                
            case 0x3:
                len += rtas_print("(Configuration error)\n");
                break;
                
            case 0x4:
                len += rtas_print("(Mainteneace procedure required)\n");
                break;
                
            case 0x9: 
                len += rtas_print("(External FRU)\n");
                break;
                
            case 0xA:
                len += rtas_print("(External Code FRU)\n");
                break;
                
            case 0xB:
                len += rtas_print("(Tool FRU)\n");
                break;
                
            case 0xC:
                len += rtas_print("(Symbolic FRU)\n");
                break;
                
            default:
                len += rtas_print("\n");
                break;
        }
    }

    if (fru->length == 4) {
        /* No optional Parts */
        len += rtas_print("\n");
        return len;
    }

    if (fru->part_no) {
        char *start = (char *)((unsigned int)fru + sizeof(*fru));

        if (fru->fru_part_no[0] != '\0') {
            len += rtas_print("%-20s", "FRU Stocking Part:");
            for (i = 0; i < 7; i++) {
                if (fru->fru_part_no == '\0')
                    break;
                len += rtas_print("%c", fru->fru_part_no[i]);
            }
            len += rtas_print("\n");
        }

        if (fru->ccin && (start[0] != '\0')) {
            len += rtas_print("%-20s", "CCIN:");
            for (i = 0; i < 4; i++) {
                if (start[i]  == '\0')
                    break;
                len += rtas_print("%c", start[i]);
            }
            len += rtas_print("\n");
            start += 4;
        }
        
        if (fru->serial_no && (start[0] != '\0')) {
            len += rtas_print("%-20s", "Serial Number:");
            for (i = 0; i < 12; i++) {
                if (start[i]  == '\0')
                    break;
                len += rtas_print("%c", start[i]);
            }
            len += rtas_print("\n");
        }

    } 
    else if (fru->maintenance && (fru->fru_procedure_id != NULL)) {
        len += rtas_print("%-20s", "Procedure ID:");
        for (i = 0; i < 5; i++) {
            if (fru->fru_procedure_id[i]  == '\0')
                break;
            len += rtas_print("%c", fru->fru_procedure_id[i]);
        }
        len += rtas_print("\n");
    }
    
    len += rtas_print("\n");
    return len;
}
    
/**
 * print_re_fru_hdr
 * @brief print the contents of an FRU section
 *
 * @param res rtas_event_scn pointer for a fru section
 * @param verbosity verbose level of output
 * @param count current fru section number that we are printing
 * @return number of bytes written
 */
int
print_re_fru_scn(struct rtas_event_scn *res, int verbosity, int count)
{
    struct rtas_v6_fru_scn *fru;
    int len = 0;

    if ((res->scn_id != RTAS_FRU_SCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    fru = (struct rtas_v6_fru_scn *)res->data;

    len += print_scn_title("FRU Section (%d)", count);

    if (verbosity >= 2) {
        len += rtas_print(PRNT_FMT_2, "Length:", fru->length, 
	                  "Call-Out Type:", fru->type);
    }

    len += rtas_print("%-20s%c - ", "Priority:", fru->priority);
    switch(fru->priority) {
        case 'H':
            len += rtas_print("High priority and mandatory call-out.\n");
            break;
            
        case 'M':
            len += rtas_print("Medium priority.\n");
            break;
            
        case 'A':
            len += rtas_print("Medium priority group A (1st group).\n");
            break;
            
        case 'B':
            len += rtas_print("Medium priority group B (2nd group).\n");
            break;
            
        case 'C':
            len += rtas_print("Medium priority group C (3rd group).\n");
            break;
            
        case 'L':
            len += rtas_print("Low Priority.\n");
            break;
    }

    if (fru->loc_code_length) {
        if (verbosity >= 2)
            len += rtas_print(PRNT_FMT_R, "Loc Code Length:", 
                              fru->loc_code_length);
        
        len += rtas_print("%-20s%s\n", "Location Code:", 
                          (char *)fru + sizeof(*fru));
    }

    if (verbosity >= 2)
        len += rtas_print("\n");

    if (res->subscn != NULL)
        len += print_re_fru_subscn(res->subscn, verbosity);

    if ((verbosity >= 2) && 
        (fru->other_platform_data || fru->other_platform_data2)) {
        char *start = (char *)((uint32_t)fru + fru->length);
        int size = ((uint32_t)fru) + fru->length - (uint32_t)start;

        if (size > 0) {
            len += rtas_print("%-20s\n", "Other Platform Data:");
            len += print_raw_data(start, size);
        }
    }

    return len;
}

/**
 * print_v6_src_subscn
 * @brief print the contents of a SRC subsection
 *
 * @param res rtas_event_scn pointer for SRC subsection
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_re_src_subscn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_src_subscn *src_sub;
    int len = 0;

    if ((res->scn_id != RTAS_SRC_SUBSCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    src_sub = (struct rtas_v6_src_subscn *)res->data;

    if (verbosity >= 2) {
        len += print_scn_title("Optional SRC Sub-Section");
        len += rtas_print(PRNT_FMT_2, "Sub-Section ID:", src_sub->id,
                          "Platform Data:", src_sub->platform_data);
        len += rtas_print(PRNT_FMT_R"\n", "Length:", src_sub->subscn_length);
    }
	
    return len;
}

/**
 * print_src_refcode
 * @brief print a detailed description of the SRC reference code
 *
 * @param src rtas_v6_src_scn pointer
 * @return number of bytes written
 */
int
print_src_refcode(struct rtas_v6_src_scn *src)
{
    int i, len = 0;

    len += rtas_print("%s \"", "Primary Reference Code:");
    for (i = 0; i < 32; i++) {
        if (src->primary_refcode[i] == '\0') 
            break;
        len += rtas_print("%c", src->primary_refcode[i]);
    }
    len += rtas_print("\"\n");

    i = 0;
    while (src_codes[i].desc != NULL) {
        if (strcmp(src->primary_refcode, src_codes[i].id) == 0) {
            len += rtas_print("%s\n", src_codes[i].desc);
            return len;
        }
        i++;
    }
    
    return len;
}

/**
 * print_re_src_scn
 * @brief print the contents of a SRC section
 *
 * @param res rtas_event_scn pointer for SRC section
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_re_src_scn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_src_scn *src;
    struct rtas_event_scn *scnp;
    int len = 0;
    int count = 1;

    if (((res->scn_id != RTAS_PSRC_SCN) && (res->scn_id != RTAS_SSRC_SCN))
        || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    src = (struct rtas_v6_src_scn *)res->data;

    if (strncmp(src->shdr.id, RTAS_PSRC_SCN_ID, 2) == 0)
        len += print_v6_scn_hdr("Primary SRC Section", &src->shdr, verbosity);
    else
        len += print_v6_scn_hdr("Secondary SRC Section", &src->shdr, verbosity);

    if (verbosity >= 2) {
        len += rtas_print(PRNT_FMT_2"\n", "SRC Version:", src->version,
                          "Subsections:", src->subscns);
    }

    len += rtas_print("Platform Data (section 1)\n");
    len += print_raw_data((char*)((uint32_t)src + sizeof(src->shdr) + 1), 1);
    
    len += rtas_print("Platform Data (section 2)\n");
    len += print_raw_data(src->platform_data, 6);
    
    len += rtas_print("\n");

    len += rtas_print("Extended Reference Codes:\n");
    len += rtas_print("2: %08x  3: %08x  4: %08x  5: %08x\n", 
                      src->ext_refcode2, src->ext_refcode3, 
                      src->ext_refcode4, src->ext_refcode5);
    len += rtas_print("6: %08x  7: %08x  8: %08x  9: %08x\n\n", 
                      src->ext_refcode6, src->ext_refcode7, 
                      src->ext_refcode8, src->ext_refcode9);

    len += print_src_refcode(src);

    len += rtas_print("\n");

    if (res->subscn != NULL)
        len += print_re_src_subscn(res->subscn, verbosity);

    scnp = res->other_scns;
    while (scnp != NULL) {
        len += print_re_fru_scn(scnp, verbosity, count);
        scnp = scnp->next;
        count++;
    }
    
    return len;
}
