/**
 * @file rtas_v6_misc.c
 * @brief Routines to print out various RTAS version 6 event sections
 *
 * Copyright (C) 2004 IBM Corporation
 * Common Public License Version 1.0 (see COPYRIGHT)
 *
 * @author Nathan Fotenot <nfont@austin.ibm.com>
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "librtasevent.h"
#include "rtas_event.h"

/**
 * months
 * @brief array of month strings
 *
 * This array is indexed wih a hex value, thats what the extra blanks
 * are for so please leave them.
 */
static char *months[] = {"", "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul",
                         "Aug", "Sep", "", "", "", "", "", "", "Oct", "Nov", 
                         "Dec"};

/**
 * print_v6_scn_hdr
 * @brief print the generic version 6 section header
 *
 * @param name section name
 * @param shdr rtas_v6_scn_hdr pointer
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_v6_scn_hdr(char *name, struct rtas_v6_scn_hdr *shdr, int verbosity)
{
    int len;

    len = print_scn_title(name);

    if (verbosity > 1) {
        len += rtas_print("%-20s      %c%c    "PRNT_FMT_R, 
                          "Section ID:", shdr->id[0], shdr->id[1], 
                          "Section Length:", shdr->length);
        len += rtas_print(PRNT_FMT_2, "Version:", shdr->version, 
                          "Sub_type:", shdr->subtype);
        len += rtas_print(PRNT_FMT_R, "Component ID:", 
                          shdr->creator_comp_id);
    }

    return len;
}

/**
 * rtas_get_main_a_scn
 * @brief retrieve the Main A section of an RTAS Event 
 *
 * @param re rtas_event pointer
 * @return pointer to rtas_event_scn on success, NULL on failure
 */
struct rtas_event_scn *
rtas_get_main_a_scn(struct rtas_event *re)
{
    return get_re_scn(re, RTAS_MAIN_A_SCN);
}

/**
 * print_re_main_a_scn
 * @brief print the RTAS version 6 Main A section
 *
 * @param res rtas_event_scn pointer to main a section
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_re_main_a_scn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_main_a_scn *maina;
    int len = 0;

    if ((res->scn_id != RTAS_MAIN_A_SCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    maina = (struct rtas_v6_main_a_scn *)res->data;

    len += print_v6_scn_hdr("Main A", &maina->shdr, verbosity);
    
    if (verbosity < 2) {
        len += rtas_print("%-20s%x %s %x\n", "Date:", maina->date.day,
                          months[maina->date.month], maina->date.year);
        len += rtas_print("%-20s%x:%x:%x:%x\n", "Time:", 
                          maina->time.hour, maina->time.minutes, 
                          maina->time.seconds, maina->time.hundredths);
    } else {
        len += rtas_print(PRNT_FMT_2, "Date:", maina->date,
                          "Time:", maina->time);
    }

    len += rtas_print("%-20s", "Creator ID:");
    switch(maina->creator_id) {
        case 'E':
	    len += rtas_print("Service Processor (%c).\n", maina->creator_id);
	    break;
            
        case 'H':
	    len += rtas_print("Hypervisor (%c).\n", maina->creator_id);
	    break;
            
        case 'W':
	    len += rtas_print("Power Control (%c).\n", maina->creator_id);
	    break;
            
        case 'L':
	    len += rtas_print("Partition Firmware (%c).\n", maina->creator_id);
	    break;
    }

    if (verbosity >= 2) {
        len += rtas_print(PRNT_FMT_2, "Section Count:", maina->scn_count, 
                          "PLID:", maina->plid);
    }
    
    len += rtas_print("\n");
    
    len += rtas_print("Platform Data (section 1):\n");
    len += print_raw_data(maina->platform_data, 8);

    len += rtas_print("Platform Data (section 2):\n");
    len += print_raw_data(maina->platform_data2, 8);

    len += rtas_print("Platform Data (section 3):\n");
    len += print_raw_data(maina->platform_data3, 4);
    len += rtas_print("\n");

    return len;
}

/**
 * rtas_get_main_b_scn
 * @brief retrieve the main b section for an RTAS event.
 *
 * @param re rtas_event pointer
 * @return rtas_event_scn pointer to main b section, NULL on failure
 */
struct rtas_event_scn *
rtas_get_main_b_scn(struct rtas_event *re)
{
    return get_re_scn(re, RTAS_MAIN_B_SCN);
}

/**
 * print_v6_main_b_subsystem_id
 * @brief Print the subsystem id from the Main B section
 *
 * @param mainb rtas_v6_main_b_scn pointer
 * @return number of bytes written
 */
int
print_v6_main_b_subsystem_id(struct rtas_v6_main_b_scn *mainb)
{
    unsigned int id = mainb->subsystem_id;
    int len = 0;

    len += rtas_print(PRNT_FMT" ", "Subsystem ID:", id);
    
    if ((id >= 0x10) && (id <= 0x1F))
        len += rtas_print("(Processor, including internal cache)\n"); 
    else if ((id >= 0x20) && (id <= 0x2F))
        len += rtas_print("(Memory, including external cache)\n");
    else if ((id >= 0x30) && (id <= 0x3F))
        len += rtas_print("(I/O (hub, bridge, bus))\n");
    else if ((id >= 0x40) && (id <= 0x4F))
        len += rtas_print("(I/O adapter, device and peripheral)\n"); 
    else if ((id >= 0x50) && (id <= 0x5F))
        len += rtas_print("(CEC Hardware)\n");
    else if ((id >= 0x60) && (id <= 0x6F))
        len += rtas_print("(Power/Cooling System)\n");
    else if ((id >= 0x70) && (id <= 0x79))
        len += rtas_print("(Other Subsystems)\n");
    else if ((id >= 0x7A) && (id <= 0x7F))
        len += rtas_print("(Surveillance Error)\n");
    else if ((id >= 0x80) && (id <= 0x8F))
        len += rtas_print("(Platform Firmware)\n");
    else if ((id >= 0x90) && (id <= 0x9F))
        len += rtas_print("(Software)\n");
    else if ((id >= 0xA0) && (id <= 0xAF))
        len += rtas_print("(External Environment)\n");
    else
        len += rtas_print("\n");

    return len;
}

/**
 * print_v6_main_b_event
 * @brief print the RTAS version 6 Main B section type data
 *
 * @param mainb rtas_v6_main_b_scn pointer
 * @return number of bytes written
 */
int
print_v6_main_b_event(struct rtas_v6_main_b_scn *mainb)
{
    int len = 0;

#define INDENT "                        "

    len += rtas_print(PRNT_FMT" ", "Severity:", mainb->event_type);
    
    if (mainb->event_type == 0) {
	len += rtas_print("Informational or non-error event,\n"INDENT);

        switch (mainb->event_subtype) {
	    case 0x01:
	        len += rtas_print("miscellaneous, informational only.\n");
		break;
	    case 0x08:
	        len += rtas_print("dump notification.\n");
		break;
	    case 0x10:
	        len += rtas_print("previously reported error has been "
                                  "corrected\n" INDENT "by system.\n");
		break;
	    case 0x20:
	        len += rtas_print("system resources manually "
                                  "deconfigured by user.\n");
		break;
	    case 0x21:
	        len += rtas_print("system resources deconfigured by "
                                  "system due to\n" INDENT 
                                  "prior error event.\n");
		break;
	    case 0x22:
	        len += rtas_print("resource deallocation event "
                                  "notification.\n");
		break;
	    case 0x30:
	        len += rtas_print("customer environmental problem has "
                                  "returned\n" INDENT "to normal.\n");
		break;
	    case 0x40:
	        len += rtas_print("concurrent maintenance event.\n");
		break;
	    case 0x60:
	        len += rtas_print("capacity upgrade event.\n");
		break;
	    case 0x70:
	        len += rtas_print("resource sparing event.\n");
		break;
	    case 0x80:
	        len += rtas_print("dynamic reconfiguration event.\n");
		break;
	    case 0xD0:
	        len += rtas_print("normal system/platform shutdown "
                                  "or powered off.\n");
		break;
	    case 0xE0:
	        len += rtas_print("platform powered off by user without "
                                  "normal\n" INDENT "shutdown.\n");
		break;
	    default:
	        len += rtas_print("unknown event subtype (%d).\n", 
                                  mainb->event_subtype);
		break;
	}
    } else {
	switch (mainb->event_type) {
	    case 0x10:
	        len += rtas_print("Recovered error, general.\n");
		break;
	    case 0x20:
	        len += rtas_print("Predictive error, general.\n");
		break;
	    case 0x21:
	        len += rtas_print("Predictive error, degraded performance.\n");
		break;
	    case 0x22:
	        len += rtas_print("Predictive error, fault may be "
                                  "corrected after\n" INDENT 
                                  "platform re-IPL.\n");
		break;
	    case 0x23:
	        len += rtas_print("Predictive Error, fault may be corrected "
                                  "after IPL,\n" INDENT 
                                  "degraded performance.\n");
		break;
	    case 0x24:
	        len += rtas_print("Predictive error, loss of redundancy.\n");
		break;
	    case 0x40:
	        len += rtas_print("Unrecoverable error, general.\n");
		break;
	    case 0x41:
	        len += rtas_print("Unrecoverable error, bypassed with "
                                  "degraded\n" INDENT "performance.\n");
		break;
	    case 0x44:
	        len += rtas_print("Unrecoverable error, bypassed with "
                                  "loss of\n" INDENT "redundancy.\n");
		break;
	    case 0x45:
	        len += rtas_print("Unrecoverable error, bypassed with "
                                  "loss of\n" INDENT "redundancy and "
                                  "performance.\n");
		break;
	    case 0x48:
	        len += rtas_print("Unrecoverable error, bypassed with "
                                  "loss of function.\n");
		break;
	    case 0x60:
	        len += rtas_print("Error on diagnostic test, general.\n");
		break;
	    case 0x61:
	        len += rtas_print("Error on diagnostic test, resource "
                                  "may produce\n" INDENT 
                                  "incorrect results.\n");
		break;
	    default:
	        len += rtas_print("Unknown event/error type (%d).\n",
                                  mainb->event_type);
		break;
	}
    }

    return len;
}

/**
 * print_v6_main_b_action
 * @brief print the RTAS version 6 Main B action data
 *
 * @param mainb rtas_v6_main_b_scn pointer
 * @return number of bytes written
 */
int
print_v6_main_b_action(struct rtas_v6_main_b_scn *mainb)
{
    int len = 0;

    len += rtas_print(PRNT_FMT" ", "Action Flag:", mainb->action);

    switch (mainb->action) {
        case 0x8000:
	    len += rtas_print("Service Action ");
	    if (mainb->action & 0x4000)
	        len += rtas_print("(hidden error) ");
	    if (mainb->action & 0x0800)
	        len += rtas_print("call home) ");
            len += rtas_print("Required.\n");
	    break;
            
	case 0x2000:
	    len += rtas_print("Report Externally, ");
	    if (mainb->action & 0x1000)
	        len += rtas_print("(HMC only).\n");
	    else
	        len += rtas_print("(HMC and Hypervisor).\n");
	    break;
            
	case 0x0400:
	    len += rtas_print("Error isolation incomplete,\n"
                              "                               "
                              "further analysis required.\n");
	    break;
            
	case 0x0000:
	    break;

	default:
	    len += rtas_print("Unknown action flag (0x%08x).\n", 
                              mainb->action);
    }

    return len;
}

/**
 * print_re_main_b_scn
 * @brief print the contents of a RTAS version 6 Main B section
 *
 * @param res rtas_event_scn ponter
 * @param verbosity verbose level of ouput
 * @return number of bytes written
 */
int
print_re_main_b_scn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_main_b_scn *mainb;
    int len = 0;

    if ((res->scn_id != RTAS_MAIN_B_SCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    mainb = (struct rtas_v6_main_b_scn *)res->data;

    len += print_v6_scn_hdr("Main B", &mainb->shdr, verbosity);
    
    len += print_v6_main_b_subsystem_id(mainb);

    len += print_v6_main_b_event(mainb);

    len += print_v6_main_b_action(mainb);

    len += rtas_print("\nPlatform Data (section 1):\n");
    len += print_raw_data((char *)&mainb->platform_data, 8);

    len += rtas_print("Platform Data (section 2):\n");
    len += print_raw_data((char *)&mainb->platform_data2, 4);
    len += rtas_print("\n");

    return len;
}

/**
 * rtas_get_mtms_scn
 * @brief retrieve the Failing Enclosure (MTMS) section of an RTAS Event
 *
 * @param re rtas_event pointer
 * @return pointer to rtas_event_scn on success, NULL on failure
 */
struct rtas_event_scn *
rtas_get_mtms_scn(struct rtas_event *re)
{
    return get_re_scn(re, RTAS_MTMS_SCN);
}

/**
 * print_re_mtms_scn
 * @brief print the contents of a Machine Type section
 *
 * @param res rtas_event_scn pointer for mtms section
 * @param verbosity verbose level of output
 * @return number of bytes written
 */
int
print_re_mtms_scn(struct rtas_event_scn *res, int verbosity)
{
    struct rtas_v6_mt_scn *mtms;
    int len = 0;
    int i;

    if ((res->scn_id != RTAS_MTMS_SCN) || (res->data == NULL)) {
        errno = EFAULT;
        return 0;
    }

    mtms = (struct rtas_v6_mt_scn *)res->data;
    
    len += print_v6_scn_hdr("Machine Type", &mtms->shdr, verbosity);

    len += rtas_print("%-20s", "Model / Type:");
    for (i = 0; i < 8; i++) {
        if (mtms->model[i] == '\0')
            break;
	len += rtas_print("%c", mtms->model[i]);
    }
    
    len += rtas_print("    %-20s", "Serial Number:");
    for (i = 0; i < 12; i++) {
        if (mtms->serial_no[i] == '\0')
            break;
	len += rtas_print("%c", mtms->serial_no[i]);
    }

    len += rtas_print("\n\n");

    return len;
}
